/*
 * This file is part of imunal
 * Copyright 2013  Guillaume Quintin, Olivier Ruatta, Philippe Gaborit
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <limits.h>

typedef unsigned long ulong;
PREFIX const ulong ulong_max = ULONG_MAX;

PREFIX ulong safe_add(ulong a,ulong b) {
  ulong r;
  if ( a == ulong_max || b == ulong_max ) return ulong_max;
  r = a + b;
  if ( r < a ) return ulong_max;
  return r;
}

PREFIX ulong safe_sub(ulong a,ulong b) {
  if ( a == ulong_max ) return ulong_max;
  if ( b > a ) return 0;
  return a - b;
}

PREFIX ulong safe_mul(ulong a,ulong b) {
  if ( a == ulong_max || b == ulong_max ) return ulong_max;
  if ( a == 0ul ) return 0ul;
  if ( b > ulong_max / a ) return ulong_max;
  return a * b;
}

PREFIX ulong safe_div(ulong a,ulong b) {
  if ( a == ulong_max ) return ulong_max;
  if ( b == ulong_max ) return 0;
  if ( b == 0ul ) return ulong_max;
  return a / b;
}

PREFIX ulong safe_sum(ulong n) {
  return safe_div(safe_mul(n,safe_add(n,1)),2);
}
